try:
    import urllib.parse as urlparse
except ImportError:
    # py2
    import urlparse

import tornado.httpclient
import tornado.web

# headers to remove as of HTTP 1.1 RFC2616
# http://www.w3.org/Protocols/rfc2616/rfc2616-sec13.html
hop_by_hop_headers = set([
    'connection',
    'keep-alive',
    'proxy-authenticate',
    'proxy-authorization',
    'te',
    'trailers',
    'transfer-encoding',
    'upgrade',
])


class ProxyHandler(tornado.web.RequestHandler):
    def __init__(self, *args, **kw):
        self.proxy_whitelist = kw.pop('proxy_whitelist', None)
        self.origin_whitelist = kw.pop('origin_whitelist', None)
        super(ProxyHandler, self).__init__(*args, **kw)

    def check_proxy_host(self, url_parts):
        if self.proxy_whitelist is None:
            return

        url = '%s://%s' % (url_parts.scheme, url_parts.netloc)

        if url in self.proxy_whitelist:
            return

        raise tornado.web.HTTPError(403)

    def check_origin(self):
        if self.origin_whitelist is None:
            return

        if 'Origin' not in self.request.headers:
            raise tornado.web.HTTPError(403)

        if self.request.headers['Origin'] not in self.origin_whitelist:
            raise tornado.web.HTTPError(403)

    def response_handler(self, response):
        if response.error and not isinstance(response.error, tornado.httpclient.HTTPError):
            self.set_status(500)
            self.write('Internal server error:\n' + str(response.error))
            self.finish()
            return

        if response.code == 599:
            # connection closed
            self.set_status(502)
            self.write('Bad gateway:\n' + str(response.error))
            self.finish()
            return

        self.set_status(response.code)
        # copy all but hop-by-hop headers
        for header, v in response.headers.items():
            if header.lower() not in hop_by_hop_headers:
                self.set_header(header, v)

        origin = self.request.headers.get('Origin', '*')
        self.set_header('Access-Control-Allow-Origin', origin)

        if self.request.method == 'OPTIONS':
            if 'Access-Control-Request-Headers' in self.request.headers:
                # allow all requested headers
                self.set_header('Access-Control-Allow-Headers',
                    self.request.headers.get('Access-Control-Request-Headers', ''))

            self.set_header('Access-Control-Allow-Methods',
                response.headers.get('Allow', ''))

            if response.code == 405:
                # fake OPTIONS response when source doesn't support it
                # as OPTIONS is required for CORS preflight requests.
                # the 405 response should contain the supported methods
                # in the Allow header.
                self.set_status(200)
                self.clear_header('Content-Length')
                self.finish()
                return

        if response.body:
            self.write(response.body)
        self.finish()

    @tornado.web.asynchronous
    def request_handler(self, url):
        url_parts = urlparse.urlparse(url)
        self.request.headers['Host'] = url_parts.netloc

        self.check_proxy_host(url_parts)
        self.check_origin()

        if self.request.query:
            url = url + '?' + self.request.query
        req = tornado.httpclient.HTTPRequest(
            url=url,
            method=self.request.method,
            body=self.request.body,
            headers=self.request.headers,
            follow_redirects=False,
            allow_nonstandard_methods=True,
            use_gzip=False, # otherwise tornado will decode proxied data
        )

        client = tornado.httpclient.AsyncHTTPClient()
        try:
            client.fetch(req, self.response_handler)
        except tornado.httpclient.HTTPError as e:
            # pass regular HTTP errors from server to client
            if hasattr(e, 'response') and e.response:
                self.response_handler(e.response)
            else:
                raise

    # alias HTTP methods to generic request handler
    SUPPORTED_METHODS = ['GET', 'POST', 'PUT', 'DELETE', 'HEAD', 'OPTIONS']
    get = request_handler
    post = request_handler
    put = request_handler
    delete = request_handler
    head = request_handler
    options = request_handler



from setuptools import setup, find_packages

setup(
    name='corsa',
    version="0.1.2",
    description='CORS proxy and web server for static apps',
    author='Oliver Tonnhofer',
    author_email='olt@bogosoft.com',
    url='https://github.com/olt/corsa',
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    entry_points = {
        'console_scripts': [
            'corsa = corsa.app:main',
        ],
    },
    install_requires=[
        'tornado',
    ],
)



from __future__ import print_function

import re
import sys

import tornado.ioloop
import tornado.web
import tornado.log

from corsa.proxy import ProxyHandler


def run_proxy(bind, app_path=None, debug=False,
    proxy_whitelist=None, origin_whitelist=None):
    handler = [
        (r'/proxy/(.*)', ProxyHandler, {
            'proxy_whitelist': proxy_whitelist,
            'origin_whitelist': origin_whitelist,
            }
        ),
    ]
    if app_path:
        handler.append(
            (r'/app/(.*)', tornado.web.StaticFileHandler, {'path': app_path})
        )
    app = tornado.web.Application(handler, debug=debug)
    app.listen(bind[1], bind[0])
    ioloop = tornado.ioloop.IOLoop.instance()
    ioloop.start()

def parse_bind_address(address, default=('localhost', 8080)):
    """
    >>> parse_bind_address('80')
    ('localhost', 80)
    >>> parse_bind_address('0.0.0.0')
    ('0.0.0.0', 8080)
    >>> parse_bind_address('0.0.0.0:8081')
    ('0.0.0.0', 8081)
    """
    if ':' in address:
        host, port = address.split(':', 1)
        port = int(port)
    elif re.match('^\d+$', address):
        host = default[0]
        port = int(address)
    else:
        host = address
        port = default[1]
    return host, port


def main():
    import optparse

    parser = optparse.OptionParser()
    parser.add_option('--app-dir')
    parser.add_option('-b', '--bind', default='localhost:8888')
    parser.add_option('--allow-proxy', default=None)
    parser.add_option('--allow-origin', default='SELF')
    parser.add_option('--debug', default=False, action='store_true')

    options, args = parser.parse_args()

    host, port = parse_bind_address(options.bind)

    tornado.log.enable_pretty_logging()

    print("Starting Corsa...", file=sys.stderr)
    print("  CORS proxy at http://%s:%d/proxy/" % (host, port), file=sys.stderr)

    if options.app_dir:
        print("  Hosting %s at http://%s:%d/app/" % (options.app_dir, host, port),
            file=sys.stderr)

    proxy_whitelist = []
    if options.allow_proxy:
        proxy_whitelist = set(options.allow_proxy.split(','))
        if 'ALL' in proxy_whitelist:
            proxy_whitelist = None

    origin_whitelist = []
    if options.allow_origin:
        origin_whitelist = set(options.allow_origin.split(','))
        if 'SELF' in origin_whitelist:
            origin_whitelist.remove('SELF')
            origin_whitelist.add('http://%s:%d' % (host, port))
        if 'ALL' in origin_whitelist:
            origin_whitelist = None

    try:
        run_proxy((host, port), options.app_dir,
            debug=options.debug,
            proxy_whitelist=proxy_whitelist,
            origin_whitelist=origin_whitelist,
        )
    except KeyboardInterrupt:
        print("Exiting", file=sys.stderr)


if __name__ == '__main__':
    main()